#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library
from datetime import datetime

# import interfaces
from onf_website.shared.common.interfaces import IWfSharedContentRoles
from pyams_security.interfaces import IRoleProtectedObject
from pyams_workflow.interfaces import IWorkflow, AUTOMATIC, IWorkflowPublicationInfo, SYSTEM, IWorkflowVersions, \
    IWorkflowState, ObjectClonedEvent, IWorkflowInfo

# import packages
from pyams_utils.registry import utility_config
from pyams_utils.request import check_request
from pyams_workflow.workflow import Transition, Workflow
from pyramid.threadlocal import get_current_registry
from zope.copy import copy
from zope.schema.vocabulary import SimpleVocabulary, SimpleTerm

from onf_website import _


#
# Workflow status
#

DRAFT = 'draft'
PROPOSED = 'proposed'
CANCELED = 'canceled'
REFUSED = 'refused'
PUBLISHED = 'published'
RETIRING = 'retiring'
RETIRED = 'retired'
ARCHIVING = 'archiving'
ARCHIVED = 'archived'
DELETED = 'deleted'

STATUS_IDS = (DRAFT,
              PROPOSED,
              CANCELED,
              REFUSED,
              PUBLISHED,
              RETIRING,
              RETIRED,
              ARCHIVING,
              ARCHIVED,
              DELETED)

VISIBLE_STATES = (PUBLISHED, RETIRING)

PROTECTED_STATES = (PUBLISHED, RETIRING, RETIRED, ARCHIVING)

READONLY_STATES = (ARCHIVED, DELETED)

WAITING_STATES = (PROPOSED, RETIRING, ARCHIVING)

STATUS_LABELS = (_("Draft"),
                 _("Proposed"),
                 _("Canceled"),
                 _("Refused"),
                 _("Published"),
                 _("Retiring"),
                 _("Retired"),
                 _("Archiving"),
                 _("Archived"),
                 _("Deleted"))

STATUS_VOCABULARY = SimpleVocabulary([SimpleTerm(STATUS_IDS[i], title=t)
                                      for i, t in enumerate(STATUS_LABELS)])


#
# Workflow conditions
#

def can_backdraft_content(wf, context):
    """Check if content can return to DRAFT state"""
    return IWorkflowPublicationInfo(context).publication_date is None


def can_propose_content(wf, content):
    """Check if a content can be proposed"""
    versions = IWorkflowVersions(content)
    return not versions.has_version(PROPOSED)


def can_retire_content(wf, context):
    """Check if already published content can return to RETIRED state"""
    return IWorkflowPublicationInfo(context).publication_date is not None


def can_create_new_version(wf, context):
    """Check if we can create a new version"""
    versions = IWorkflowVersions(context)
    return not (versions.has_version(DRAFT) or
                versions.has_version(PROPOSED) or
                versions.has_version(CANCELED) or
                versions.has_version(REFUSED))


#
# Workflow actions
#

def publish_action(wf, context):
    """Publish version"""
    IWorkflowPublicationInfo(context).publication_date = datetime.utcnow()
    translate = check_request().localizer.translate
    version_id = IWorkflowState(context).version_id
    for version in IWorkflowVersions(context).get_versions((PUBLISHED, RETIRING, RETIRED, ARCHIVING)):
        if version is not context:
            IWorkflowInfo(version).fire_transition_toward('archived',
                                                          comment=translate(_("Published version {0}")).format(version_id))


def archive_action(wf, context):
    """Remove readers when a content is archived"""
    roles = IWfSharedContentRoles(context, None)
    if roles is not None:
        IRoleProtectedObject(context).revoke_role('onf.Reader', roles.readers)


def clone_action(wf, context):
    """Create new version"""
    result = copy(context)
    registry = get_current_registry()
    registry.notify(ObjectClonedEvent(result, context))
    return result


def delete_action(wf, context):
    """Delete draft version, and parent if single version"""
    versions = IWorkflowVersions(context)
    versions.remove_version(IWorkflowState(context).version_id)


#
# Workflow transitions
#

init = Transition(transition_id='init',
                  title=_("Initialize"),
                  source=None,
                  destination=DRAFT)

draft_to_proposed = Transition(transition_id='draft_to_proposed',
                               title=_("Propose publication"),
                               source=DRAFT,
                               destination=PROPOSED,
                               condition=can_propose_content,
                               permission='onf.ManageContent',
                               menu_css_class='fa fa-fw fa-question',
                               view_name='wf-propose.html',
                               order=1)

retired_to_proposed = Transition(transition_id='retired_to_proposed',
                                 title=_("Propose publication"),
                                 source=RETIRED,
                                 destination=PROPOSED,
                                 condition=can_propose_content,
                                 permission='onf.ManageContent',
                                 menu_css_class='fa fa-fw fa-question',
                                 view_name='wf-propose.html',
                                 order=1)

proposed_to_canceled = Transition(transition_id='proposed_to_canceled',
                                  title=_("Cancel publication request"),
                                  source=PROPOSED,
                                  destination=CANCELED,
                                  permission='onf.ManageContent',
                                  menu_css_class='fa fa-fw fa-mail-reply',
                                  view_name='wf-cancel-propose.html',
                                  order=2)

canceled_to_draft = Transition(transition_id='canceled_to_draft',
                               title=_("Reset canceled publication to draft"),
                               source=CANCELED,
                               destination=DRAFT,
                               trigger=AUTOMATIC,
                               condition=can_backdraft_content)

canceled_to_retired = Transition(transition_id='canceled_to_retired',
                                 title=_("Reset canceled publication to retired"),
                                 source=CANCELED,
                                 destination=RETIRED,
                                 trigger=AUTOMATIC,
                                 condition=can_retire_content)

proposed_to_refused = Transition(transition_id='proposed_to_refused',
                                 title=_("Refuse publication"),
                                 source=PROPOSED,
                                 destination=REFUSED,
                                 permission='onf.PublishContent',
                                 menu_css_class='fa fa-fw fa-thumbs-o-down',
                                 view_name='wf-refuse.html',
                                 order=3)

refused_to_draft = Transition(transition_id='refused_to_draft',
                              title=_("Reset refused publication to draft"),
                              source=REFUSED,
                              destination=DRAFT,
                              trigger=AUTOMATIC,
                              condition=can_backdraft_content)

refused_to_retired = Transition(transition_id='refused_to_retired',
                                title=_("Reset refused publication to retired"),
                                source=REFUSED,
                                destination=RETIRED,
                                trigger=AUTOMATIC,
                                condition=can_retire_content)

proposed_to_published = Transition(transition_id='proposed_to_published',
                                   title=_("Publish content"),
                                   source=PROPOSED,
                                   destination=PUBLISHED,
                                   permission='onf.PublishContent',
                                   action=publish_action,
                                   menu_css_class='fa fa-fw fa-thumbs-o-up',
                                   view_name='wf-publish.html',
                                   order=4)

published_to_retiring = Transition(transition_id='published_to_retiring',
                                   title=_("Request retiring"),
                                   source=PUBLISHED,
                                   destination=RETIRING,
                                   permission='onf.ManageContent',
                                   menu_css_class='fa fa-fw fa-pause',
                                   view_name='wf-retiring.html',
                                   order=7)

retiring_to_published = Transition(transition_id='retiring_to_published',
                                   title=_("Cancel retiring request"),
                                   source=RETIRING,
                                   destination=PUBLISHED,
                                   permission='onf.ManageContent',
                                   menu_css_class='fa fa-fw fa-mail-reply',
                                   view_name='wf-cancel-retiring.html',
                                   order=8)

retiring_to_retired = Transition(transition_id='retiring_to_retired',
                                 title=_("Retire content"),
                                 source=RETIRING,
                                 destination=RETIRED,
                                 permission='onf.PublishContent',
                                 menu_css_class='fa fa-fw fa-stop',
                                 view_name='wf-retire.html',
                                 order=9)

retired_to_archiving = Transition(transition_id='retired_to_archiving',
                                  title=_("Request archive"),
                                  source=RETIRED,
                                  destination=ARCHIVING,
                                  permission='onf.ManageContent',
                                  menu_css_class='fa fa-fw fa-archive',
                                  view_name='wf-archiving.html',
                                  order=10)

archiving_to_retired = Transition(transition_id='archiving_to_retired',
                                  title=_("Cancel archiving request"),
                                  source=ARCHIVING,
                                  destination=RETIRED,
                                  permission='onf.ManageContent',
                                  menu_css_class='fa fa-fw fa-mail-reply',
                                  view_name='wf-cancel-archiving.html',
                                  order=11)

archiving_to_archived = Transition(transition_id='archiving_to_archived',
                                   title=_("Archive content"),
                                   source=ARCHIVING,
                                   destination=ARCHIVED,
                                   permission='onf.ManageContent',
                                   action=archive_action,
                                   menu_css_class='fa fa-fw fa-archive',
                                   view_name='wf-archive.html',
                                   order=12)

published_to_archived = Transition(transition_id='published_to_archived',
                                   title=_("Archive published content"),
                                   source=PUBLISHED,
                                   destination=ARCHIVED,
                                   trigger=SYSTEM,
                                   action=archive_action)

retiring_to_archived = Transition(transition_id='retiring_to_archived',
                                  title=_("Archive retiring content"),
                                  source=RETIRING,
                                  destination=ARCHIVED,
                                  trigger=SYSTEM,
                                  action=archive_action)

retired_to_archived = Transition(transition_id='retired_to_archived',
                                 title=_("Archive retired content"),
                                 source=RETIRED,
                                 destination=ARCHIVED,
                                 trigger=SYSTEM,
                                 action=archive_action)

published_to_draft = Transition(transition_id='published_to_draft',
                                title=_("Create new version"),
                                source=PUBLISHED,
                                destination=DRAFT,
                                permission='onf.CreateContent',
                                condition=can_create_new_version,
                                action=clone_action,
                                menu_css_class='fa fa-fw fa-file-o',
                                view_name='wf-clone.html',
                                order=13)

retiring_to_draft = Transition(transition_id='retiring_to_draft',
                               title=_("Create new version"),
                               source=RETIRING,
                               destination=DRAFT,
                               permission='onf.CreateContent',
                               condition=can_create_new_version,
                               action=clone_action,
                               menu_css_class='fa fa-fw fa-file-o',
                               view_name='wf-clone.html',
                               order=14)

retired_to_draft = Transition(transition_id='retired_to_draft',
                              title=_("Create new version"),
                              source=RETIRED,
                              destination=DRAFT,
                              permission='onf.CreateContent',
                              condition=can_create_new_version,
                              action=clone_action,
                              menu_css_class='fa fa-fw fa-file-o',
                              view_name='wf-clone.html',
                              order=15)

archiving_to_draft = Transition(transition_id='archiving_to_draft',
                                title=_("Create new version"),
                                source=ARCHIVING,
                                destination=DRAFT,
                                permission='onf.CreateContent',
                                condition=can_create_new_version,
                                action=clone_action,
                                menu_css_class='fa fa-fw fa-file-o',
                                view_name='wf-clone.html',
                                order=16)

archived_to_draft = Transition(transition_id='archived_to_draft',
                               title=_("Create new version"),
                               source=ARCHIVED,
                               destination=DRAFT,
                               permission='onf.CreateContent',
                               condition=can_create_new_version,
                               action=clone_action,
                               menu_css_class='fa fa-fw fa-file-o',
                               view_name='wf-clone.html',
                               order=17)

delete = Transition(transition_id='delete',
                    title=_("Delete version"),
                    source=DRAFT,
                    destination=DELETED,
                    action=delete_action,
                    permission='onf.ManageContent',
                    menu_css_class='fa fa-fw fa-trash',
                    view_name='wf-delete.html',
                    order=18)

wf_transitions = [init,
                  draft_to_proposed,
                  retired_to_proposed,
                  proposed_to_canceled,
                  canceled_to_draft,
                  canceled_to_retired,
                  proposed_to_refused,
                  refused_to_draft,
                  refused_to_retired,
                  proposed_to_published,
                  published_to_retiring,
                  retiring_to_published,
                  retiring_to_retired,
                  retired_to_archiving,
                  archiving_to_retired,
                  published_to_archived,
                  retiring_to_archived,
                  retired_to_archived,
                  archiving_to_archived,
                  published_to_draft,
                  retiring_to_draft,
                  retired_to_draft,
                  archiving_to_draft,
                  archived_to_draft,
                  delete]

wf = Workflow(wf_transitions,
              states=STATUS_VOCABULARY,
              published_states=VISIBLE_STATES)


@utility_config(name='ONF website workflow', provides=IWorkflow)
class WorkflowUtility(object):
    """ONF website workflow utility"""

    def __new__(cls):
        return wf
